/*
 * Unit-tests against MPFQ
 */
#include <unistd.h>
#include <syscall.h>
#include <stdlib.h>
#include "minunit.h"
#include "mpfq.h"
#include "rbc.h"
#include "utils.h"

#define VERBOSE 0


/*
 * Variables
 */

int INSTANCES = 10;

mpfq_field field1;
mpfq_elt a1;
mpfq_elt b1;
mpfq_elt c1;
mpfq_elt_ur d1;

rbc_elt a2;
rbc_elt b2;
rbc_elt c2;
rbc_elt_ur d2;



/*
 * Initialization and destruction
 */
void test_setup() {

  // MPFQ initialization
  mpfq_field_init(field1);
  mp_limb_t base = 2;
  mpfq_field_setopt(field1, MPFQ_IO_TYPE, &base);

  mpfq_init(field1, a1);
  mpfq_init(field1, b1);
  mpfq_init(field1, c1);
  mpfq_elt_ur_init(field1, d1);

  mpfq_set_random(field1, a1);
  mpfq_set_random(field1, b1);
  mpfq_set_zero(field1, c1);
  mpfq_set_random2(field1, d1);

  // RBC initialization
  rbc_field_init();
  rbc_elt_set_from_uint64(a2, a1);
  rbc_elt_set_from_uint64(b2, b1);
  rbc_elt_set_from_uint64(c2, c1);
  rbc_elt_ur_set_from_uint64(d2, d1);

  #if VERBOSE
    printf("\na1: "); mpfq_elt_print_uint64(a1);
    printf("\na2: "); rbc_elt_print(a2);
    printf("\n");
    printf("\nb1: "); mpfq_elt_print_uint64(b1);
    printf("\nb2: "); rbc_elt_print(b2);
    printf("\n");
  #endif
}

void test_teardown() {};

void MU_SUITE_CLEAR() {
  mpfq_clear(field1, a1);
  mpfq_clear(field1, b1);
  mpfq_clear(field1, c1);
  mpfq_field_clear(field1);
}



/*
 * Test functions
 */
MU_TEST(test_elt_add) {
  mpfq_add(field1, c1, b1, a1);
  rbc_elt_add(c2, a2, b2);

  #if VERBOSE
    printf("\nc1: "); mpfq_elt_print_uint64(c1);
    printf("\nc2: "); rbc_elt_print(c2);
    printf("\n");
  #endif

  mu_assert(rbc_elt_equal_to_mpfq(c1, c2) == 1, "Failure: test_elt_add");
}



MU_TEST(test_elt_mul) {
  mpfq_mul(field1, c1, a1, b1);
  rbc_elt_mul(c2, a2, b2);

  #if VERBOSE
    printf("\nc1: "); mpfq_elt_print_uint64(a1);
    printf("\nc2: "); rbc_elt_print(a2);
    printf("\n");
  #endif

  mu_assert(rbc_elt_equal_to_mpfq(c1, c2) == 1, "Failure: test_elt_mul");
}



MU_TEST(test_elt_sqr) {
  mpfq_sqr(field1, c1, a1);
  rbc_elt_sqr(c2, a2);

  #if VERBOSE
    printf("\na1: "); mpfq_elt_print_uint64(a1);
    printf("\na2: "); rbc_elt_print(a2);
    printf("\n");
    printf("\nc1: "); mpfq_elt_print_uint64(c1);
    printf("\nc2: "); rbc_elt_print(c2);
    printf("\n");
  #endif

  mu_assert(rbc_elt_equal_to_mpfq(c1, c2) == 1, "Failure: test_elt_sqr");
}



MU_TEST(test_elt_inv) {
  mpfq_inv(field1, c1, a1);
  rbc_elt_inv(c2, a2);

  #if VERBOSE
    printf("\nc1: "); mpfq_elt_print_uint64(c1);
    printf("\nc2: "); rbc_elt_print(c2);
    printf("\n");
  #endif

  mu_assert(rbc_elt_equal_to_mpfq(c1, c2) == 1, "Failure: test_elt_inv");
}



MU_TEST(test_elt_red) {
  mpfq_reduce(field1, c1, d1);
  rbc_elt_reduce(c2, d2);

  #if VERBOSE
    printf("\nd1: "); mpfq_elt_ur_print_uint64(d1);
    printf("\nd2: "); rbc_elt_ur_print(d2);
    printf("\n");
    printf("\nc1: "); mpfq_elt_print_uint64(c1);
    printf("\nc2: "); rbc_elt_print(c2);
    printf("\n");
  #endif

  mu_assert(rbc_elt_equal_to_mpfq(c1, c2) == 1, "Failure: test_elt_red");
}



/*
 * Test execution
 */
MU_TEST_SUITE(test_suite) {
  MU_SUITE_CONFIGURE(&test_setup, &test_teardown);
  MU_RUN_TEST(test_elt_add);
  MU_RUN_TEST(test_elt_mul);
  MU_RUN_TEST(test_elt_sqr);
  MU_RUN_TEST(test_elt_inv);
  MU_RUN_TEST(test_elt_red);
  MU_SUITE_CLEAR();
}

int main() {
  for(int i = 0 ; i < INSTANCES ; ++i) {
    MU_RUN_SUITE(test_suite);
  }

  MU_REPORT();
  return minunit_fail;
}

